# The structure of the dataset, posterior and trainer classes are based on the structures used in ScVI
# The implementation of ScVI by YosefLab is available on GitHub

import numpy as np
import pandas as pd
import scipy.sparse
from torch.utils.data import Dataset
import scanpy

from typing import Union, Dict, List
from collections import OrderedDict, defaultdict
import logging

logger = logging.getLogger(__name__)


class GeneExpressionDataset(Dataset):

    def __init__(self):

        # registers
        self.gene_attribute_names = set()
        self.cell_attribute_names = set()
        self.attribute_mappings = defaultdict(list)

        # attributes
        self._data = None
        self._raw = None
        self._batch_indices = None
        self._labels = None
        self.num_batches = None
        self.num_labels = None
        self.gene_names = None
        self.cell_types = None

    def __repr__(self): # todo finish this function
        pass

    def from_data(
            self,
            data: Union[np.ndarray, scipy.sparse.csr_matrix],
            raw: Union[np.ndarray, scipy.sparse.csr_matrix] = None,
            batch_indices: Union[List[int], np.ndarray, scipy.sparse.csr_matrix] = None,
            labels: Union[List[int], np.ndarray, scipy.sparse.csr_matrix] = None,
            gene_names: Union[List[str], np.ndarray] = None,
            cell_types: Union[List[str], np.ndarray] = None,
            cell_attributes_dict: Dict[str, Union[np.ndarray, List]] = None,
            gene_attributes_dict: Dict[str, Union[np.ndarray, List]] = None
                ):
        # todo write description of parameters

        # data = data.T
        self._data = (
            np.ascontiguousarray(data, dtype=np.float32)
            if isinstance(data, np.ndarray)
            else data
        )

        self._raw = (
            np.ascontiguousarray(raw, dtype=np.float32)
            if isinstance(raw, np.ndarray)
            else raw
        )

        self.initialize_cell_attribute(
            "batch_indices",
            np.asarray(batch_indices).reshape((-1, 1))
            if batch_indices is not None
            else np.zeros((data.shape[0], 1))
        )

        self.initialize_cell_attribute(
            "labels",
            np.ndarray(labels).reshape((-1, 1))
            if labels is not None
            else np.zeros((data.shape[0], 1))
        )

        if gene_names is not None:
            genes = np.asarray(gene_names, dtype="<U64")
            self.initialize_gene_attribute("gene_name", genes)
            if len(np.unique(self.gene_names)) != len(self.gene_names):
                logger.warning("Gene names are not unique.")

        if cell_types is not None:
            self.initialize_mapped_attribute(
                "labels", "cell_types", np.asarray(cell_types, dtype = "<U128")
            )
        elif labels is None:
            self.initialize_mapped_attribute(
                "labels", "cell_types", np.asarray(["undefined"], dtype = "<U128")
            )

        if cell_attributes_dict:
            for attribute_name, attribute_value in cell_attributes_dict.items():
                self.initialize_cell_attribute(attribute_name, attribute_value)
        if gene_attributes_dict:
            for attribute_name, attribute_value in gene_attributes_dict.items():
                self.initialize_gene_attribute(attribute_name, attribute_value)

    def initialize_cell_attribute(self, attribute_name, attribute):

        try:
            len_attribute = attribute.shape[0]
        except AttributeError:
            len__attribute = len(attribute)

        if not self.nb_cells == len_attribute:
            raise ValueError(
                "Number of cells ({number_cells}) and length of attribute ({number_attributes})"
                "are not the same.".format(number_cells=self.nb_cells, number_attributes=len_attribute)
            )
        setattr(
            self,
            attribute_name,
            np.asarray(attribute)
            if not isinstance(attribute, scipy.sparse.csr_matrix)
            else attribute
        )
        self.cell_attribute_names.add(attribute_name)

    def initialize_gene_attribute(self, attribute_name, attribute_value):
        if not self.nb_genes == len(attribute_value):
            raise ValueError(
                "Number of genes ({n_genes}) and length of gene attribute ({n_attr}) mismatch".format(
                    n_genes=self.nb_genes, n_attr=len(attribute_value)
                )
            )
        setattr(self, attribute_name, attribute_value)
        self.gene_attribute_names.add(attribute_name)

    def initialize_mapped_attribute(self, source_attribute_name, mapped_name, mapped_value):
        source_attribute = getattr(self, source_attribute_name)
        if isinstance(source_attribute, np.ndarray):
            source_type = source_attribute.dtype
        else:
            source_type = type(source_attribute[0])
        if not np.issubdtype(source_type, np.integer):
            raise ValueError(
                "The mapping (source) attribute ({source_name}) should be categorical "
                "and hence integer and not ({given_type})".format(source_name=source_attribute_name, given_type=source_type)
            )
        max_categorical_value = np.max(source_attribute)
        if not max_categorical_value <= len(mapped_value):
            raise ValueError(
                "The maximum categorical value ({max_val}) is less than mapped values ({len_mapped})".format(
                    max_val=max_categorical_value, len_mapped=len(mapped_value)
                )
            )
        self.attribute_mappings[source_attribute_name].append(mapped_name)
        setattr(self, mapped_name, mapped_value)

    # REGISTERS AND PROPERTIES

    def __len__(self):
        return self.nb_cells

    def __getitem__(self, item):
        return item

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data: Union[np.ndarray, scipy.sparse.csr_matrix, scipy.sparse.coo_matrix]):
        num_dim = len(data.shape)
        if num_dim != 2:
            raise ValueError(
                "Gene Expression data should have 2 dimensions and not {}".format(num_dim)
            )
        if type(data) == scipy.sparse.coo_matrix:
            data = scipy.sparse.csr_matrix(data)
        self._data = data

    @property
    def raw(self):
        return self._raw

    @raw.setter
    def raw(self, raw: Union[np.ndarray, scipy.sparse.csr_matrix, scipy.sparse.coo_matrix]):
        num_dim = len(raw.shape)
        if num_dim != self._data.size:
            raise ValueError(
                "Gene Expression raw data should have matching dimensions as data."
            )
        if type(raw) == scipy.sparse.coo_matrix:
            raw = scipy.sparse.csr_matrix(raw)
        self._raw = raw

    @property
    def nb_cells(self) -> int:
        return self.data.shape[0]

    @property
    def nb_genes(self) -> int:
        return self.data.shape[1]

    @property
    def nb_cell_counts(self) -> np.ndarray:
        return np.sum(self.data, axis=1)

    @property
    def batch_indices(self) -> np.ndarray:
        return self._batch_indices

    @batch_indices.setter
    def batch_indices(self, batch_indices: Union[List[int], np.ndarray]):
        batch_indices = np.asarray(batch_indices, dtype=np.int64).reshape((-1, 1))
        self.num_batches = len(np.unique(batch_indices))
        self._batch_indices = batch_indices

    @property
    def labels(self) -> np.ndarray:
        return self._labels

    @labels.setter
    def labels(self, labels: Union[List[int], np.ndarray]):
        labels = np.asarray(labels, dtype=np.int64)
        self.num_labels = len(np.unique(labels))
        self._labels = labels


def normalize(
        adata: scanpy.AnnData,
        filter_min_counts=None,
        size_factors=False,
        scale_input=False,
        logtrans_input=False
):

    if filter_min_counts:
        scanpy.pp.filter_genes(adata, min_counts=filter_min_counts)
        scanpy.pp.filter_cells(adata, min_counts=filter_min_counts)

    adata.raw = adata
    dataset = GeneExpressionDataset()
    dataset.from_data(adata.X, raw=adata.raw.X)

    if size_factors:
        scanpy.pp.normalize_per_cell(adata)
        size_factor_cell = dataset.nb_cell_counts / np.median(dataset.nb_cell_counts)
        dataset.initialize_cell_attribute("size_factor", size_factor_cell.reshape((-1,1)))
    if logtrans_input:
        scanpy.pp.log1p(adata)
    if scale_input:
        scanpy.pp.scale(adata)

    dataset.data = adata.X
    obs_index = adata.obs.columns
    for i, index in enumerate(obs_index):
        dataset.initialize_cell_attribute(index, adata.obs.iloc[:,i].values)
    return dataset





import numpy as np
from typing import List
import torch
from torch import nn
import torch.optim as optim
import torch.nn.functional as F


from scdeep.trainer import Trainer
from scdeep.dataset import GeneExpressionDataset
from scdeep.network import AutoEncoder, LinearActivation


class scScope(AutoEncoder):
    def __init__(
            self,
            input_d,
            encoder_layers_dim: List,
            decoder_layers_dim: List,
            latent_layer_out_dim: int,
            exp_batch_input,
            t: int = 2
    ):
        self.t = t
        self.input_dim = input_d.shape[1]

        super(scScope, self).__init__(input_d, encoder_layers_dim, decoder_layers_dim, latent_layer_out_dim,
                                      activation='relu', weight_initializer='normal', weight_init_params={'std': 0.1},
                                      bias_initializer='zeros', batchnorm=False)

        if len(exp_batch_input) > 0:
            num_batch = exp_batch_input.shape[1]
        else:
            num_batch = 1

        self.batch_effect_layer = nn.Linear(num_batch, self.input_dim, bias=False)
        nn.init.zeros_(self.batch_effect_layer.weight)

        impute_layer1 = LinearActivation(self.input_dim, 64, activation='relu',
                                         weight_init='normal', weight_init_params={'std': 0.1},
                                         bias_init='zeros')

        impute_layer2 = LinearActivation(64, self.input_dim, activation=None,
                                         weight_init='normal', weight_init_params={'std': 0.1},
                                         bias_init='zeros')

        # impute_layer2 = nn.Linear(64, self.input_dim)
        # nn.init.zeros_(impute_layer2.bias)
        # nn.init.normal_(impute_layer2.weight, mean=0.0, std=0.1)

        self.imputation_model = nn.Sequential(impute_layer1, impute_layer2)

    def forward(self, x, exp_batch_input):
        self.input = x
        latent_features_list = []
        output_list = []
        batch_effect_removal_layer = self.batch_effect_layer(exp_batch_input)
        for i in range(self.t):
            if i == 0:
                x = F.relu(x - batch_effect_removal_layer)
            else:
                imputed = self.imputation_model(output)
                imputed = torch.mul(1 - torch.sign(self.input), imputed)
                x = F.relu(imputed + self.input - batch_effect_removal_layer)
            encoded = self.encode(x)
            latent_features = self.latent_layer(encoded)
            decoded = self.decode(latent_features)
            output = self.output_layer(decoded)
            output_list.append(output)
            latent_features_list.append(latent_features)

        return output_list, latent_features_list, batch_effect_removal_layer


class scScopeTrainer(Trainer):

    def __init__(
            self,
            model: scScope,
            gene_dataset: GeneExpressionDataset,
            use_mask: bool = True,
            **kwargs
    ):
        super().__init__(model, gene_dataset, **kwargs)

        self.use_mask = use_mask

        if self.gene_dataset.num_batches > 1:
            self.one_hot_batches = np.zeros((self.gene_dataset.nb_cells, self.gene_dataset.num_batches))
            self.one_hot_batches[:, (self.gene_dataset.batch_indices.reshape((1, -1)) - 1)] = 1.0
        else:
            self.one_hot_batches = np.zeros_like(self.gene_dataset.batch_indices)
        self.one_hot_batches = torch.tensor(self.one_hot_batches, dtype=torch.float32, device=self.device)
        # self.one_hot_batches = self.one_hot_batches.cuda() if self.use_cuda else self.one_hot_batches

        (train_set, test_set, val_set) = self.train_test_validation()
        self.register_posterior(train_set, test_set, val_set)

    def on_training_begin(self):
        self.optimizer = optim.Adam(self.model.parameters(), lr=self.lr)
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                print(name, param.shape)

    def model_output(self, data_tensor):

        data, indices = data_tensor
        batch_data = self.one_hot_batches[indices.long(), :]

        output_list, latent_list, batch_effect_removal_layer = self.model(data, batch_data)
        return output_list, latent_list, batch_effect_removal_layer, data

    def on_training_loop(self, data_tensor):
        output_list, latent_list, batch_effect_removal_layer, data = self.model_output(data_tensor)
        self.current_loss = loss = self.loss(output_list, data, use_mask=self.use_mask,
                                             batch_effect_removal_layer=batch_effect_removal_layer)
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

    @torch.no_grad()
    def on_validation(self, data_tensor, loss):
        output_list, latent_list, batch_effect_removal_layer, data = self.model_output(data_tensor)
        loss.append(self.loss(output_list, data, use_mask=self.use_mask,
                              batch_effect_removal_layer=batch_effect_removal_layer).item())
        return loss

    def loss(self, output_layer_list, input_d, use_mask, batch_effect_removal_layer):
        # input_d_corrected = input_d - batch_effect_removal_layer
        # if use_mask:
        #     val_mask = torch.sign(input_d_corrected)
        # else:
        #     val_mask = torch.sign(input_d_corrected + 1)

        # for i in range(len(output_layer_list)):
        #     out_layer = output_layer_list[i]
        #     if i == 0:
        #         loss_value = torch.norm(torch.mul(val_mask, (out_layer - input_d_corrected))) / \
        #                      torch.norm(torch.mul(val_mask, input_d))
        #     else:
        #         loss_value += torch.norm(torch.mul(val_mask, (out_layer - input_d_corrected))) / \
        #                       torch.norm(torch.mul(val_mask, input_d))
        # return loss_value
        input_d_corrected = input_d - batch_effect_removal_layer
        val_mask = torch.sign(input_d_corrected)
        for i in range(len(output_layer_list)):
            out_layer = output_layer_list[i]
            if i == 0:
                loss_value = (torch.norm(torch.mul(val_mask, out_layer - input_d_corrected)))
            else:
                loss_value = loss_value + (torch.norm(torch.mul(val_mask, out_layer - input_d_corrected)))
        return loss_value

    @torch.no_grad()
    def predict(self, input_d, batch_data):

        if len(np.unique(batch_data)) > 1:
            one_hot_batches = np.zeros((batch_data.shape[0], len(np.unique(batch_data))))
            one_hot_batches[:, (batch_data.reshape((1, -1)) - 1)] = 1
        else:
            one_hot_batches = np.zeros_like(batch_data)
        input_d = torch.tensor(input_d, device=self.device)
        one_hot_batches = torch.tensor(one_hot_batches, dtype=torch.float32, device=self.device)

        output_layer, latent_layer, batch_removal_layer = self.model(input_d,
                                                                     one_hot_batches)

        output_layer_numpy = [o.cpu().detach().numpy() for o in output_layer]
        latent_layer_numpy = [l.cpu().detach().numpy() for l in latent_layer]
        batch_removal_layer_numpy = batch_removal_layer.cpu().detach().numpy()

        return output_layer_numpy, latent_layer_numpy, batch_removal_layer_numpy




import numpy as np
from typing import List
import torch
from torch import nn
import torch.optim as optim
from tqdm import tqdm
from sklearn.metrics import adjusted_rand_score, normalized_mutual_info_score
from sklearn.cluster import KMeans


from scdeep.trainer import Trainer
from scdeep.dataset import GeneExpressionDataset
from scdeep.network import ZINBAutoEncoder
from scdeep.losses import zinb_loss


def cluster_acc(y_true, y_pred):
    """
    Calculate clustering accuracy. Require scikit-learn installed
    # Arguments
        y: true labels, numpy.array with shape `(n_samples,)`
        y_pred: predicted labels, numpy.array with shape `(n_samples,)`
    # Return
        accuracy, in [0,1]
    """
    y_true = y_true.astype(np.int64)
    assert y_pred.size == y_true.size
    D = max(y_pred.max(), y_true.max()) + 1
    w = np.zeros((D, D), dtype=np.int64)
    for i in range(y_pred.size):
        w[y_pred[i], y_true[i]] += 1
    from sklearn.utils.linear_assignment_ import linear_assignment
    ind = linear_assignment(w.max() - w)
    return sum([w[i, j] for i, j in ind]) * 1.0 / y_pred.size


class ClusteringLayer(nn.Module):

    def __init__(
            self,
            input_dim,
            n_clusters: int,
            alpha=1.0,
            initial_weights=None,
    ):
        super(ClusteringLayer, self).__init__()
        self.n_features = input_dim
        self.n_clusters = n_clusters
        self.alpha = alpha
        self.initial_weights = initial_weights

        if self.initial_weights is None:
            weights = torch.tensor((self.n_clusters, self.n_features))
            nn.init.xavier_uniform_(weights)
            self.clusters = nn.Parameter(weights)
        else:
            self.clusters = nn.Parameter(initial_weights)

    def forward(self, x):
        """ student t-distribution, as same as used in t-SNE algorithm.
            q_ij = 1/(1+dist(x_i, u_j)^2), then normalize it.
        Arguments:
            x: the variable containing data, shape=(n_samples, n_features)
        Return:
            q: student's t-distribution, or soft labels for each sample. shape=(n_samples, n_clusters)
        """
        x = (torch.sum(torch.square(torch.unsqueeze(x, dim=1) - self.clusters), dim=2) / self.alpha)
        q = 1.0 / (1.0 + x)
        q = torch.pow(q, ((self.alpha + 1.0) / 2.0))
        q = (q.T / torch.sum(q, dim=1)).T
        return q


class scDeepCluster(ZINBAutoEncoder):

    def __init__(
            self,
            input_d,
            n_clusters: int,
            encoder_layers_dim: List,
            decoder_layers_dim: List,
            latent_layer_out_dim: int,
            activation='relu',
            weight_initializer='xavier',
            noise_sd: float = 0.,
            **kwargs
    ):
        super(scDeepCluster, self).__init__(input_d, encoder_layers_dim, decoder_layers_dim, latent_layer_out_dim,
                                            batchnorm=False, activation=activation,
                                            weight_initializer=weight_initializer, **kwargs)
        self.input_dim = input_d.shape[1]
        self.n_clusters = n_clusters
        self.pretrain = False
        self.noise_sd = noise_sd
        self.clustering_layer = None

    def pretrain_over(self, initial_cluster_weights):
        self.pretrain = True
        self.clustering_layer = ClusteringLayer(self.input_dim, self.n_clusters, initial_weights=initial_cluster_weights)

    def forward(self, x, size_factors):
        if not self.pretrain:
            noise = ((self.noise_sd ** 0.5)*torch.randn(size=x.shape))
            noise = x.new_tensor(noise)
            x = x + noise
            for encoder_module in self.encode:
                x = encoder_module.linear(x)
                noise = ((self.noise_sd ** 0.5)*torch.randn(size=x.shape))
                noise = x.new_tensor(noise.detach())
                x = x + noise
                x = encoder_module.act_layer(x)
            latent_output = self.latent_layer(x)
            x = self.decode(latent_output)
            mean = self.output_layer(x)
            mean = mean * size_factors.reshape((-1, 1))
            theta = self.theta_layer(x)
            pi = self.pi_layer(x)
            return latent_output, [mean, theta, pi]
        else:
            encoded = self.encode(x)
            latent_output = self.latent_layer(encoded)
            decoded = self.decode(latent_output)
            clustering_output = self.clustering_layer(latent_output)
            mean = self.output_layer(decoded)
            mean = mean * size_factors.reshape((-1, 1))
            pi = self.pi_layer(decoded)
            theta = self.theta_layer(decoded)
            return clustering_output, [mean, theta, pi]


class scDeepClusterTrainer(Trainer):

    def __init__(
            self,
            model: scDeepCluster,
            gene_dataset: GeneExpressionDataset,
            use_mask: bool = True,
            scale_factor=1.0,
            ridge_lambda=0.0,
            l1_coeff=0.,
            l2_coeff=0.,
            **kwargs
    ):
        super(scDeepClusterTrainer, self).__init__(model, gene_dataset, **kwargs)

        self.use_mask = use_mask
        self.eps = 1e-10
        self.scale_factor = scale_factor
        self.l1_coeff = l1_coeff
        self.l2_coeff = l2_coeff
        self.ridge_lambda = ridge_lambda
        self.update_interval = None
        self.tol = None
        self.loss_weights = None
        self.latent_output = None

        (train_set, test_set, val_set) = self.train_test_validation()
        self.register_posterior(train_set, test_set, val_set)

        self.y_pred = []

    def pretrain(self, n_epochs=200, lr=0.001, ae_file='ae_weights'):

        self.model.pretrain = False
        self.loss = zinb_loss
        self.optimizer = optim.Adam(self.model.parameters(), lr=lr)

        self.model.train()
        for self.epoch in tqdm(
                range(n_epochs),
                desc="pre-training",
                disable=not self.show_progressbar
        ):
            self.num_iter = 0
            print("Epoch: {}".format(self.epoch + 1))
            for data_tensor in self.data_load_loop("train_set"):
                data, indices = data_tensor

                size_factors_layer = data.new_tensor(self.gene_dataset.size_factor[indices, :])
                #size_factors_layer = size_factors_layer.cuda() if self.use_cuda else size_factors_layer

                latent_output, output = self.model(data, size_factors_layer)
                raw_data = data.new_tensor(self.gene_dataset.raw[indices, :])
                self.current_loss = loss = self.loss(raw_data, output, eps=self.eps, scale_factor=self.scale_factor,
                                                     ridge_lambda=self.ridge_lambda)
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
                self.on_iteration_end()

        self.save_checkpoint(ae_file)
        self.model.pretrain = True
        self.model.eval()

    @torch.no_grad()
    def extract_feature(self, x):  # extract features from before clustering layer
        x = x.to(self.device)
        encoded = self.model.encode(x)
        latent_output = self.model.latent_layer(encoded)
        return latent_output

    @torch.no_grad()
    def predict_clusters(self, x):  # predict cluster labels using the output of clustering layer
        q = self.model.clustering_layer(x)
        return torch.argmax(q, dim=1)

    @torch.no_grad()
    def target_distribution(self, q):  # target distribution P which enhances the discrimination of soft label Q
        self.model.eval()
        weight = q ** 2 / torch.sum(q, dim=0)
        weight = (weight.T / torch.sum(weight, dim=1)).T
        return weight

    def training_extras_init(self, n_clusters, ae_file, update_interval=20, tol=1e-3, loss_weights=(1, 1)):
        self.update_interval = update_interval
        self.tol = tol
        self.loss_weights = loss_weights
        if not self.model.pretrain and not ae_file:
            print('...pretraining model with default hyperparamters')
            self.pretrain(ae_file=ae_file)
        elif ae_file is not None:
            self.load_checkpoint(ae_file)
            print('...autoencoder weights loaded successfully')
        self.epoch = -1

        self.model.eval()
        print('Initializing cluster centers with k-means.')
        kmeans = KMeans(n_clusters=n_clusters, n_init=20)
        self.y_pred = kmeans.fit_predict(self.extract_feature(torch.tensor(self.gene_dataset.data, device=self.device)).cpu().detach().numpy())
        self.y_pred_last = np.copy(self.y_pred)
        self.model.pretrain_over(torch.tensor(kmeans.cluster_centers_, device=self.device))
        self.model.train()

    def on_training_begin(self):
        torch.autograd.set_detect_anomaly(True)
        self.optimizer = optim.Adadelta(self.model.parameters(), lr=self.lr)
        self.kl_loss = torch.nn.KLDivLoss(reduction='mean')
        self.loss = zinb_loss

        for name, param in self.model.named_parameters():
            if param.requires_grad:
                print(name, param.shape)

    def on_epoch_begin(self):
        self.num_iter = 0
        print("Epoch: {}".format(self.epoch + 1))
        self.model.eval()
        if self.epoch % self.update_interval == 0 or self.epoch == self.num_epochs - 1:
            self.q, _ = self.model(torch.tensor(self.gene_dataset.data, device=self.device),
                                   torch.tensor(self.gene_dataset.size_factor, device=self.device))
            self.p = self.target_distribution(self.q).detach()

            self.y_pred = torch.argmax(self.q, dim=1)
            self.y_pred = self.y_pred.cpu().detach().numpy()
            # check accuracy of algorithm so far
            if self.gene_dataset.labels is not None:
                acc = np.round(cluster_acc(self.gene_dataset.labels, self.y_pred), 8)
                ari = np.round(adjusted_rand_score(self.gene_dataset.labels, self.y_pred), 8)
                nmi = np.round(normalized_mutual_info_score(self.gene_dataset.labels, self.y_pred), 8)
                print("Iter: {}\nACC: {:.4f}\n ARI: {:.4f}\nNMI: {:.4f}\nLoss: {:.4f}\n\n".format(self.epoch, acc, ari,
                                                                                                  nmi, self.current_loss))

            # check stop criterion
            delta_label = np.sum(self.y_pred != self.y_pred_last).astype(np.float32) / self.y_pred.shape[0]
            self.y_pred_last = np.copy(self.y_pred)
            if self.epoch > 0 and delta_label < self.tol:
                print('delta_label ', delta_label, '< tol ', self.tol)
                print('Reached tolerance threshold...')
                # todo figure out how to stop training
        self.model.train()

    def on_training_loop(self, data_tensor):
        clustering_output, output = self.model_output(data_tensor)
        data, indices = data_tensor
        raw_data = data.new_tensor(self.gene_dataset.raw[indices, :])
        self.optimizer.zero_grad()
        loss1 = self.loss(raw_data, output, eps=self.eps, scale_factor=self.scale_factor, ridge_lambda=self.ridge_lambda)
        loss2 = self.kl_loss(clustering_output.log(), self.p[indices.long(), :])
        #print('KL Loss: {:.4f}\nAE Loss: {:.4f}\n'.format(loss2.item(), loss1.item()))
        loss = self.current_loss = self.loss_weights[0]*loss1 + self.loss_weights[1]*loss2

        loss.backward()
        self.optimizer.step()
        # for name, param in self.model.named_parameters():
        #     if name == "clustering_layer.clusters":
        #         print(param.grad)
        #         print(param.grad.shape)
        #         print(param)

    def model_output(self, data_tensor):
        data, indices = data_tensor
        size_factors_layer = data.new_tensor(self.gene_dataset.size_factor[indices, :])
        # size_factors_layer = size_factors_layer.cuda() if self.use_cuda else size_factors_layer

        clustering_output, output = self.model(data, size_factors_layer)
        return clustering_output, output

    def on_epoch_end(self):
        pass

    def on_training_end(self):
        pass

    @torch.no_grad()
    def predict(self, x):
        x = torch.tensor(x, device=self.device)
        latent_output = self.extract_feature(x)
        clusters = self.predict_clusters(latent_output)
        latent_numpy = latent_output.cpu().detach().numpy()
        clusters_numpy = clusters.cpu().detach().numpy()

        return latent_numpy, clusters_numpy





import numpy as np
import pandas as pd
from typing import List, Union, DefaultDict, Dict
import logging
import torch
from torch import nn


class ExponentialActivation(nn.Module):

    def __init__(self):
        super(ExponentialActivation, self).__init__()

    def forward(self, x):
        x = torch.exp(x)
        return x


class LinearActivation(nn.Module):

    activations = {'relu': nn.ReLU(), 'sigmoid': nn.Sigmoid(), 'softmax': nn.Softmax(),
                   'exp': ExponentialActivation(), 'softplus': nn.Softplus()}
    initializers = {'xavier': nn.init.xavier_uniform_, 'zeros': nn.init.zeros_, 'normal': nn.init.normal_}

    def __init__(
            self,
            input_dim,
            out_dim,
            batchnorm=False,
            activation='relu',
            dropout=0.,
            weight_init=None,
            weight_init_params: Dict = None,
            bias_init=None,
            bias_init_params: Dict = None
    ):
        super(LinearActivation, self).__init__()
        if weight_init_params is None:
            weight_init_params = {}
        if bias_init_params is None:
            bias_init_params = {}
        self.batchnorm_layer = None
        self.act_layer = None
        self.dropout_layer = None
        self.linear = nn.Linear(input_dim, out_dim)

        if weight_init is not None:
            self.initializers[weight_init](self.linear.weight, **weight_init_params)
        if bias_init is not None:
            self.initializers[bias_init](self.linear.bias, **bias_init_params)

        if batchnorm:
            self.batchnorm_layer = nn.BatchNorm1d(num_features=out_dim)
        if activation:
            self.act_layer = self.activations[activation]
        if dropout > 0.0:
            self.dropout_layer = nn.Dropout(dropout)

    def forward(self, x):
        x = self.linear(x)
        if self.batchnorm_layer:
            x = self.batchnorm_layer(x)
        if self.act_layer:
            x = self.act_layer(x)
        if self.dropout_layer:
            x = self.dropout_layer(x)
        return x


class AutoEncoder(nn.Module):
    def __init__(
            self,
            input_d,
            encoder_layers_dim: List,
            decoder_layers_dim: List,
            latent_layer_out_dim: int,
            batchnorm: bool = True,
            activation: str = 'relu',
            dropout: float = 0.,
            weight_initializer=None,
            weight_init_params: Dict = None,
            bias_initializer=None,
            bias_init_params: Dict = None
    ):
        super(AutoEncoder, self).__init__()
        if weight_init_params is None:
            weight_init_params = {}
        if bias_init_params is None:
            bias_init_params = {}
        self.input_dim = input_d.shape[1]
        self.batchnorm = batchnorm

        encode_layers = []
        if len(encoder_layers_dim) > 0:
            for i in range(len(encoder_layers_dim)):
                if i == 0:
                    encode_layers.append(LinearActivation(self.input_dim, encoder_layers_dim[i],
                                                          batchnorm=self.batchnorm, activation=activation,
                                                          dropout=dropout,
                                                          weight_init=weight_initializer,
                                                          weight_init_params=weight_init_params,
                                                          bias_init=bias_initializer, bias_init_params=bias_init_params))
                else:
                    encode_layers.append(LinearActivation(encoder_layers_dim[i - 1], encoder_layers_dim[i],
                                                          batchnorm=self.batchnorm, activation=activation,
                                                          dropout=dropout,
                                                          weight_init=weight_initializer,
                                                          weight_init_params=weight_init_params,
                                                          bias_init=bias_initializer, bias_init_params=bias_init_params))
            self.latent_layer_input_dim = encoder_layers_dim[-1]
        else:
            self.latent_layer_input_dim = self.input_dim
        self.encode = nn.Sequential(*encode_layers)

        self.latent_layer = LinearActivation(self.latent_layer_input_dim, latent_layer_out_dim,
                                             batchnorm=self.batchnorm, activation=activation,
                                             dropout=dropout,
                                             weight_init=weight_initializer, weight_init_params=weight_init_params,
                                             bias_init=bias_initializer, bias_init_params=bias_init_params)

        decode_layers = []
        if len(decoder_layers_dim) > 0:
            for i in range(len(decoder_layers_dim)):
                if i == 0:
                    decode_layers.append(LinearActivation(latent_layer_out_dim, decoder_layers_dim[i],
                                                          batchnorm=self.batchnorm, activation=activation,
                                                          dropout=dropout,
                                                          weight_init=weight_initializer,
                                                          weight_init_params=weight_init_params,
                                                          bias_init=bias_initializer, bias_init_params=bias_init_params))
                else:
                    decode_layers.append(LinearActivation(decoder_layers_dim[i - 1], decoder_layers_dim[i],
                                                          batchnorm=self.batchnorm, activation=activation,
                                                          dropout=dropout,
                                                          weight_init=weight_initializer,
                                                          weight_init_params=weight_init_params,
                                                          bias_init=bias_initializer, bias_init_params=bias_init_params))
            self.output_layer_input_dim = decoder_layers_dim[-1]
        else:
            self.output_layer_input_dim = latent_layer_out_dim
        self.decode = nn.Sequential(*decode_layers)

        self.output_layer = LinearActivation(self.output_layer_input_dim, self.input_dim,
                                             activation=activation,
                                             weight_init=weight_initializer, weight_init_params=weight_init_params,
                                             bias_init=bias_initializer, bias_init_params=bias_init_params)

    def forward(self, x):
        encoded = self.encode(x)
        latent = self.latent_layer(encoded)
        decoded = self.decode(latent)
        output = self.output_layer(decoded)

        return latent, output


class NBAutoEncoder(AutoEncoder):
    def __init__(
            self,
            input_d,
            encoder_layers_dim: List,
            decoder_layers_dim: List,
            latent_layer_out_dim: int,
            batchnorm: bool = True,
            activation: str = 'relu',
            weight_initializer='xavier',
            **kwargs
    ):
        self.batchnorm = batchnorm
        super(NBAutoEncoder, self).__init__(input_d, encoder_layers_dim=encoder_layers_dim,
                                            decoder_layers_dim=decoder_layers_dim, latent_layer_out_dim=latent_layer_out_dim,
                                            batchnorm=batchnorm, activation=activation,
                                            weight_initializer=weight_initializer, **kwargs)

        # self.latent_layer = LinearActivation(self.latent_layer_input_dim, latent_layer_out_dim,
        #                                      batchnorm=self.batchnorm, activation=None, weight_init=weight_initializer)
        # the self.output_layer will be equivalent to the mean layer
        self.output_layer = LinearActivation(self.output_layer_input_dim, self.input_dim,
                                             activation='exp', weight_init=weight_initializer)

        self.theta_layer = LinearActivation(self.output_layer_input_dim, self.input_dim,
                                            activation='softplus', weight_init=weight_initializer)

    def forward(self, x, size_factors):
        encoded = self.encode(x)
        latent_output = self.latent_layer(encoded)
        decoded = self.decode(latent_output)
        mean = self.output_layer(decoded)
        mean = mean * size_factors.reshape((-1, 1))
        theta = self.theta_layer(decoded)
        return latent_output, [mean, theta]


class ZINBAutoEncoder(AutoEncoder):
    def __init__(
            self,
            input_d,
            encoder_layers_dim: List,
            decoder_layers_dim: List,
            latent_layer_out_dim: int,
            batchnorm: bool = True,
            activation: str = 'relu',
            weight_initializer='xavier',
            **kwargs
    ):
        self.batchnorm = batchnorm
        super(ZINBAutoEncoder, self).__init__(input_d, encoder_layers_dim=encoder_layers_dim,
                                              decoder_layers_dim=decoder_layers_dim, latent_layer_out_dim=latent_layer_out_dim,
                                              batchnorm=batchnorm, activation=activation,
                                              weight_initializer=weight_initializer, **kwargs)

        # self.latent_layer = LinearActivation(self.latent_layer_input_dim, latent_layer_out_dim,
        #                                      batchnorm=self.batchnorm, activation=None, weight_init=weight_initializer)
        # the self.output_layer will be equivalent to the mean layer
        self.output_layer = LinearActivation(self.output_layer_input_dim, self.input_dim,
                                             activation='exp', weight_init=weight_initializer)

        self.theta_layer = LinearActivation(self.output_layer_input_dim, self.input_dim,
                                            activation='softplus', weight_init=weight_initializer)

        self.pi_layer = LinearActivation(self.output_layer_input_dim, self.input_dim,
                                         activation='sigmoid', weight_init=weight_initializer)

    def forward(self, x, size_factors):
        encoded = self.encode(x)
        latent_output = self.latent_layer(encoded)
        decoded = self.decode(latent_output)
        mean = self.output_layer(decoded)
        mean = mean * size_factors.reshape((-1, 1))
        theta = self.theta_layer(decoded)
        pi = self.pi_layer(decoded)
        return latent_output, [mean, theta, pi]



# DeepImpute implementation in PyTorch

import numpy as np
import pandas as pd
import scipy.sparse
import scipy.io
import torch
import torch.nn as nn
import torch.nn.functional as F
from scipy.stats import expon
import torch.optim as optim
from typing import Union

from scdeep.trainer import Trainer
from scdeep.dataset import GeneExpressionDataset


class SubModule(nn.Module):
    def __init__(self, input_dim, output_dim=512):
        super(SubModule, self).__init__()
        self.dense_layer = nn.Linear(input_dim, 256)
        self.dropout = nn.Dropout(0.2)
        self.output_layer = nn.Linear(256, output_dim)

    def forward(self, x):
        x = F.relu(self.dense_layer(x))
        x = self.dropout(x)
        x = F.softplus(self.output_layer(x))
        return x


class DeepImputeModel(nn.Module):

    def __init__(self):
        super(DeepImputeModel, self).__init__()
        self.inp = []
        self.module_list = nn.ModuleList([])

    def initialize_modules(self, input_dims):
        self.inp = input_dims
        for i in self.inp:
            self.module_list.append(SubModule(input_dim=i))

    def forward(self, x):
        # x = torch.split(x, self.inp, dim=1)
        output = []
        for i, mod in enumerate(self.module_list):
            output.append(mod(x[i]))
        output = torch.cat(output, dim=1)
        return output


class DeepImputeTrainer(Trainer):

    def __init__(
            self,
            model: DeepImputeModel,
            gene_dataset: GeneExpressionDataset,
            genes_to_impute=None,
            min_vmr=0.5,
            nn_lim="auto",
            number_predictor=None,
            n_top_correlated_genes=5,
            subset_dim=512,
            mode="random",
            **kwargs
    ):
        super().__init__(model, gene_dataset, **kwargs)

        self.targets = []
        self.predictors = []
        self.min_vmr = min_vmr
        self.subset_dim = subset_dim

        if type(self.gene_dataset.data) == np.ndarray or type(self.gene_dataset.data) == scipy.sparse.csr_matrix:
            self.gene_dataset.data = pd.DataFrame(self.gene_dataset.data, dtype='float32')
        else:
            self.gene_dataset.data = pd.DataFrame(self.gene_dataset.data.values, dtype='float32')
        # (note: below operations are carried out on data assuming it is a pandas dataframe)
        genes_vmr = (self.gene_dataset.data.var() / (1 + self.gene_dataset.data.mean())).sort_values(ascending=False)
        genes_vmr = genes_vmr[genes_vmr > 0]

        # In case 1, while filling genes, we repeat genes that have been previously selected
        # but do not choose genes that have a VMR < 0.5
        # In case 2, irrespective of VMR, we select genes from the already selected gene pool for filling genes
        if genes_to_impute is None:
            genes_to_impute = self.filter_genes(genes_vmr, min_vmr, nn_lim)
        else:
            number_genes = len(genes_to_impute)
            if number_genes % self.subset_dim != 0:
                number_fill = self.subset_dim - (number_genes % self.subset_dim)
                fill_genes = genes_to_impute[:number_fill]
                genes_to_impute = np.concatenate((genes_to_impute, fill_genes))

        self.corr_matrix = self.correlation_matrix(self.gene_dataset.data, number_predictor)

        # the  next two functions save the INDICES of genes that are to form the predictors and targets
        self.set_targets(self.gene_dataset.data.reindex(columns=genes_to_impute), mode)
        self.set_predictors(self.corr_matrix, n_top_correlated_genes)

        self.gene_dataset.data = np.log1p(self.gene_dataset.data.values).astype(np.float32)
        # self.gene_dataset.data = torch.tensor(self.gene_dataset.data.values)

        (train_set, test_set, val_set) = self.train_test_validation()
        self.register_posterior(train_set, test_set, val_set)

    def loss(self, y_pred, y_true):
        l = []
        for i, ytrue in enumerate(y_true):
            a = ytrue - y_pred[i]
            y = (ytrue * torch.mul(a, a))
            l.append(y.mean())
        return l

    def on_training_begin(self):
        self.model.initialize_modules([len(pred) for pred in self.predictors])
        self.model = self.model.cuda() if self.use_cuda else self.model
        for name, param in self.model.named_parameters():
            print(name)
            print(param.shape)
        self.optimizer = optim.Adam(self.model.parameters(), lr=self.lr)

    def model_output(self, data_tensor):
        data_tensors, indices = data_tensor
        inp = [data_tensors[:, column] for column in self.predictors]
        # inp = torch.split(data_tensors, [len(i) for i in self.predictors], dim=1)
        output = self.model(inp)
        output = torch.split(output, [len(t) for t in self.targets], dim=1)
        target = [data_tensors[:, column] for column in self.targets]
        return output, target

    def on_training_loop(self, data_tensor):
        output, target = self.model_output(data_tensor)
        self.current_loss = loss = self.loss(output, target)
        self.current_loss = torch.mean(torch.stack(loss))
        self.optimizer.zero_grad()
        for l in loss:
            l.backward(retain_graph=True)
        self.optimizer.step()

    @torch.no_grad()
    def on_validation(self, data_tensor, loss):
        output, target = self.model_output(data_tensor)
        loss.append(np.asarray([l.item() for l in self.loss(output, target)]).mean())
        return loss

    def correlation_matrix(self, data, number_predictor=None):

        # we find CV to find one value per gene to calculate the correlation between two genes
        # data = data.loc[:, ~data.columns.duplicated()]
        cv = data.std() / data.mean()
        cv[np.isinf(cv)] = 0
        if number_predictor is None:
            predictions = data.loc[:, cv > 0]
        else:
            predictions = data.loc[:, cv.sort_values(ascending=False).index[:number_predictor]]
        # the strongest linear relationship between two values
        # is represented by the absolute value of the Pearson's correlation coefficient
        corr_mat = pd.DataFrame(np.abs(np.corrcoef(predictions, rowvar=False)), index=predictions.columns,
                                columns=predictions.columns).fillna(0)
        return corr_mat

    def filter_genes(self, genes, min_vmr: Union[int, float], nn_lim):
        if not str(nn_lim).isdigit():
            nn_lim = (genes > min_vmr).sum()

        number_subsets = int(np.ceil(nn_lim / self.subset_dim))
        genes_to_impute = genes.index[:number_subsets * self.subset_dim]

        rest = self.subset_dim - (len(genes_to_impute) % self.subset_dim)
        if rest > 0 and rest != self.subset_dim:
            fill_genes = np.random.choice(genes.index, rest)
            genes_to_impute = np.concatenate((genes_to_impute, fill_genes))
        # genes_to_impute contains the indices of genes that should be included for imputation
        return genes_to_impute

    def set_targets(self, data, mode):  # mode = random / progressive

        number_subsets = int(data.shape[1] / self.subset_dim)
        if mode == 'progressive':
            self.targets = data.columns.values.reshape((number_subsets, self.subset_dim))
        else:
            self.targets = np.random.choice(data.columns, size=(number_subsets, self.subset_dim), replace=False)
        # self.targets = list(self.targets)

    def set_predictors(self, covariance_matrix, n_top_correlated_genes):
        self.predictors = []
        for i, target in enumerate(self.targets):
            predictor = covariance_matrix.loc[target].drop(np.intersect1d(target, covariance_matrix.columns), axis=1)
            sorted_args = np.argsort(-predictor.values, axis=1)[:, :n_top_correlated_genes]
            predictor = predictor.columns[sorted_args.flatten()]

            self.predictors.append(np.unique(predictor))

            print("Network {}: {} predictors, {} targets".format(i, len(np.unique(predictor)), len(target)))

    def get_probs(self, vec, distr):
        return {
            "exp": expon.pdf(vec, 0, 20),
            "uniform": np.tile([1. / len(vec)], len(vec)),
        }.get(distr)

    def mask_data(self, data_to_mask, test_size, distr="exp", dropout=0.01):
        np.random.seed(self.seed)

        permuted_indices = np.random.permutation(data_to_mask.shape[0])
        test_set = int(np.ceil(data_to_mask.shape[0] * test_size))
        test_indices = np.array(permuted_indices[:test_set])

        data_to_mask = data_to_mask[test_indices, :]
        bin_mask = np.ones(data_to_mask.shape).astype(bool)

        for c in range(data_to_mask.shape[0]):
            cells = data_to_mask[c, :]
            positive_indices = np.arange(data_to_mask.shape[1])[cells > 0]
            positive_cells = cells[positive_indices]

            if positive_cells.size > 5:
                probs = self.get_probs(positive_cells, distr)
                n_masked = 1 + int(dropout * len(positive_cells))
                if n_masked >= positive_cells.size:
                    print("Warning: too many cells masked")
                    n_masked = 1 + int(0.5 * len(positive_cells))
                masked_idx = np.random.choice(
                    positive_cells.size, n_masked, p=probs / probs.sum(), replace=False
                )
                bin_mask[c, positive_indices[sorted(masked_idx)]] = False
        unmasked_dataset = np.copy(data_to_mask)
        data_to_mask[~bin_mask] = 0
        return unmasked_dataset, data_to_mask, test_indices

    @torch.no_grad()
    def predict(self, data, return_imputed_only=False, policy="restore"):

        data_tensor = torch.tensor(data.values, device=self.device)

        inp = [data_tensor[:, column] for column in self.predictors]

        predicted = self.model(inp)

        predicted = pd.DataFrame(predicted.cpu().detach().numpy(), index=data.index, columns=self.targets.flatten())
        predicted = predicted.groupby(by=predicted.columns, axis=1).mean()
        not_predicted = data.drop(columns=self.targets.flatten())
        imputed = pd.concat([predicted, not_predicted], axis=1).loc[data.index, data.columns]

        if policy == "restore":
            not_replaced_values = (data > 0)
            imputed[not_replaced_values] = data[not_replaced_values]
        elif policy == "max":
            not_replaced_values = (data > imputed)
            imputed[not_replaced_values] = data[not_replaced_values]

        return imputed

    def model_score(self, data, test_size=0.2):

        data, masked_dataset, row_indices = self.mask_data(data, test_size=test_size)
        data = pd.DataFrame(data, index=row_indices)
        masked_dataset = pd.DataFrame(masked_dataset, index=row_indices)

        imputed = self.predict(masked_dataset)

        imputed_genes = np.intersect1d(data.columns, imputed.columns)
        masked_index = data[imputed_genes].values != masked_dataset[imputed_genes].values

        score = self.metric_score(data[imputed_genes].values[masked_index], imputed[imputed_genes].values[masked_index])
        return score

    def metric_score(self, a, b):
        return scipy.stats.pearsonr(a.reshape(-1), b.reshape(-1))[0]










from typing import Union
import torch
from torch import nn
import torch.optim as optim


from scdeep.trainer import Trainer
from scdeep.dataset import GeneExpressionDataset
from scdeep.network import NBAutoEncoder, ZINBAutoEncoder
from scdeep.losses import nb_loss, zinb_loss


class DCATrainer(Trainer):
    def __init__(
            self,
            model: Union[NBAutoEncoder, ZINBAutoEncoder],
            gene_dataset: GeneExpressionDataset,
            use_mask: bool = True,
            scale_factor=1.0,
            ridge_lambda=0.0,
            l1_coeff=0.,
            l2_coeff=0.,
            **kwargs
    ):
        super().__init__(model, gene_dataset, **kwargs)

        self.use_mask = use_mask
        self.eps = 1e-10
        self.scale_factor = scale_factor
        self.l1_coeff = l1_coeff
        self.l2_coeff = l2_coeff
        self.ridge_lambda = None

        (train_set, test_set, val_set) = self.train_test_validation()
        self.register_posterior(train_set, test_set, val_set)

        if type(self.model).__name__ == 'ZINBAutoEncoder':
            self.ridge_lambda = ridge_lambda
            self.loss = zinb_loss
        else:
            self.loss = nb_loss
        self.latent_output = None

    def on_training_begin(self):
        self.optimizer = optim.RMSprop(self.model.parameters(), lr=self.lr)

        for name, param in self.model.named_parameters():
            if param.requires_grad:
                print(name, param.shape)

    def model_output(self, data_tensor):
        data, indices = data_tensor

        size_factors_layer = data.new_tensor(self.gene_dataset.size_factor[indices, :])
        # size_factors_layer = size_factors_layer.cuda() if self.use_cuda else size_factors_layer

        latent_output, output = self.model(data, size_factors_layer)
        self.latent_output = latent_output
        return output, data

    def on_training_loop(self, data):
        output, data_tensors = self.model_output(data)
        self.optimizer.zero_grad()

        self.current_loss = loss = self.loss(data_tensors, output, eps=self.eps, scale_factor=self.scale_factor,
                                             ridge_lambda=self.ridge_lambda)
        l1_reg = torch.tensor(0., device=self.device)
        l2_reg = torch.tensor(0., device=self.device)
        for param in self.model.parameters():
            l1_reg += torch.norm(param, p=1)
            l2_reg += (torch.norm(param, p=2) ** 2)
        self.current_loss = loss = loss + self.l1_coeff*l1_reg + self.l2_coeff*l2_reg

        loss.backward()
        nn.utils.clip_grad_value_(self.model.parameters(), clip_value=5)
        self.optimizer.step()

    @torch.no_grad()
    def predict(self, input_d, size_factor):
        latent_output, output = self.model(torch.tensor(input_d, device=self.device),
                                           torch.tensor(size_factor, device=self.device))
        latent_output = latent_output.cpu().detach().numpy()
        output = [l.cpu().detach().numpy() for l in output]
        return latent_output, output



import setuptools

VERSION = "0.0.1"

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scdeep",
    version=VERSION,
    author="BayesLabs",
    author_email="contact@bayeslabs.co",
    description="single-cell analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bayeslabs/Deepscan",
    packages=setuptools.find_packages(),
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


# The structure of the dataset, posterior and trainer classes are based on the structures used in ScVI
# The implementation of ScVI by YosefLab is available on GitHub

import logging
import numpy as np
import torch
import copy

from torch.utils.data import DataLoader, SubsetRandomSampler
from scdeep.dataset import GeneExpressionDataset

logger = logging.getLogger(__name__)


class Posterior:

    def __init__(
            self,
            model,
            gene_dataset: GeneExpressionDataset,
            shuffle=True,
            indices=None,
            use_cuda=True,
            data_loader_kwargs=dict()
    ):
        self.model = model
        self.gene_dataset = gene_dataset
        self.use_cuda = use_cuda

        if indices is not None:
            self.sampler = SubsetRandomSampler(indices)
        else:
            self.sampler = None

        self.data_loader_kwargs = copy.deepcopy(data_loader_kwargs)
        self.data_loader_kwargs.update({"sampler": self.sampler})
        self.data_loader = DataLoader(gene_dataset, **self.data_loader_kwargs)

    @property
    def num_cells(self):
        # if hasattr(self.data_loader.sampler, "indices"):
        #     return len(self.data_loader.sampler.indices)
        # else:
        #   return len(self.gene_dataset)
        return len(self.gene_dataset)

    @property
    def indices(self) -> np.ndarray:
        # if hasattr(self.data_loader.sampler, "indices"):
        #     return self.data_loader.sampler.indices
        # else:
        #   return np.arange(len(self.gene_dataset))
        return np.arange(len(self.gene_dataset))

    def __len__(self):
        return self.num_cells

    def __iter__(self):
        return map(self.to_cuda, [(torch.tensor(self.gene_dataset.data[i,:]),i) for i in iter(self.data_loader)])

    def to_cuda(self, t):
        return t[0].cuda() if self.use_cuda else t[0], t[1]










import torch


from scdeep.utils import nan2inf, reduce_mean


def nb_loss(y_true, output, mean=True, eps=1e-10, scale_factor=1.0, ridge_lambda=None, mask=False):
    y_pred, theta = output

    y_true = y_true.float()
    y_pred = y_pred.float() * scale_factor
    y_pred = y_pred + 1e-6
    zeros = theta.new_zeros(size=theta.shape) + 1e6
    theta = torch.min(theta, zeros)
    t1 = torch.lgamma(theta + eps) + torch.lgamma(y_true + 1.0) - torch.lgamma(y_true + theta + eps)
    t2 = (theta + y_true) * torch.log(1.0 + (y_pred / (theta + eps))) + (
            y_true * (torch.log(theta + eps) - torch.log(y_pred + eps)))

    loss = t1 + t2

    loss = nan2inf(loss)

    if mean:
        if mask:
            loss = reduce_mean(loss)
        else:
            loss = torch.mean(loss)

    return loss


def zinb_loss(y_true, output, mean=True, eps=1e-10, scale_factor=1.0, ridge_lambda=0., mask=False):
    y_pred, theta, pi = output

    nb_case = nb_loss(y_true, [y_pred, theta], mean=False, eps=eps, scale_factor=scale_factor) - torch.log(1.0 - pi + eps)

    y_true = y_true.float()
    y_pred = y_pred.float() * scale_factor
    zeros = theta.new_zeros(size=theta.shape) + 1e6
    theta = torch.min(theta, zeros)

    zero_nb = torch.pow(theta / (theta + y_pred + eps), theta)
    zero_case = -(torch.log(pi + ((1.0 - pi) * zero_nb) + eps))
    result = torch.where((y_true < 1e-8), zero_case, nb_case)
    ridge = ridge_lambda * torch.square(pi)
    result += ridge

    result = nan2inf(result)

    if mean:
        if mask:
            result = reduce_mean(result)
        else:
            result = torch.mean(result)
    return result



# The structure of the dataset, posterior and trainer classes are based on the structures used in ScVI
# The implementation of ScVI by YosefLab is available on GitHub

import logging
import time
from typing import Union, List
from collections import OrderedDict
import os

import numpy as np
import torch
import matplotlib.pyplot as plt
from statistics import mean

from tqdm import tqdm

from scdeep.posterior import Posterior
from scdeep.dataset import GeneExpressionDataset

logger = logging.getLogger(__name__)

# todo change the train size and test size functions and add a predict function instead and a test function

class Trainer:
    default_metrics_to_monitor = []

    def __init__(
            self,
            model,
            gene_dataset: GeneExpressionDataset,
            use_cuda: bool = True,
            metrics_to_monitor: List = None,
            frequency_stats: int = None,
            weight_decay: float = None,
            data_loader_kwargs: dict = None,
            show_progressbar: bool = True,
            batch_size: int = 128,
            train_size: Union[int, float] = 0.8,
            test_size: Union[int, float] = None,
            shuffle: bool = True,
            seed: int = 0,
            max_nans: int = 15,
            output_dir=None
    ):

        # model and dataset
        self.model = model
        self.gene_dataset = gene_dataset
        self.seed = seed
        self.use_cuda = torch.cuda.is_available() and use_cuda
        if self.use_cuda:
            self.model = self.model.cuda()
        self.device = torch.device('cuda') if self.use_cuda else torch.device('cpu')

        # dataloader
        self._posterior = OrderedDict()
        self.batch_size = batch_size
        self.data_loader_kwargs = {"batch_size": batch_size, "pin_memory": use_cuda}
        data_loader_kwargs = data_loader_kwargs if data_loader_kwargs else dict()
        self.data_loader_kwargs.update(data_loader_kwargs)

        # optimization
        self.optimizer = None
        self.weight_decay = weight_decay
        self.num_epochs = None
        self.lr = None
        self.epoch = -1
        self.num_iter = 0
        self.training_time = 0

        # training nans
        self.max_nans = max_nans
        self.current_loss = None
        self.was_previous_loss_nan = False
        self.nan_count = 0
        self.training_loss_per_epoch = []
        self.training_loss = [[], []]
        self.validation_loss = [[], []]

        self.output_dir = output_dir
        self.train_size = train_size
        self.test_size = test_size
        self.shuffle = shuffle
        self.frequency_stats = 1 if frequency_stats is None else frequency_stats
        # todo metrics and early stopping

        self.show_progressbar = show_progressbar
        self.compute_metrics_time = 0

    def train(self, num_epochs=500, lr=1e-3, params=None, **extra_kwargs):

        begin = time.time()
        self.model.train()

        # initialization of other model's parameters
        self.training_extras_init(**extra_kwargs)

        self.compute_metrics_time = 0
        self.num_epochs = num_epochs
        self.lr = lr

        # todo compute_metrics function if needed

        self.on_training_begin()

        for self.epoch in tqdm(
            range(num_epochs),
            desc="training",
            disable=not self.show_progressbar
        ):
            self.on_epoch_begin()

            for data_tensor in self.data_load_loop("train_set"):
                self.on_iteration_begin()
                self.on_training_loop(data_tensor)
                self.on_iteration_end()
            self.on_epoch_end()

        # todo compute_metrics and early stopping and saving best metrics

        self.model.eval()
        self.training_extras_end()

        self.training_time += (time.time() - begin) - self.compute_metrics_time
        self.on_training_end()

    def model_output(self, data_tensor):
        data, indices = data_tensor
        output = self.model(data)
        return output

    def on_training_loop(self, data_tensor):
        output, data = self.model_output(data_tensor)
        self.current_loss = loss = self.loss(data, output)
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

    def training_extras_init(self, **extra_kwargs):
        pass

    def training_extras_end(self):
        pass

    def on_training_begin(self):
        pass

    def on_epoch_begin(self):
        self.num_iter = 0
        print("\nEpoch: {}\n".format(self.epoch + 1))
        self.training_loss_per_epoch = []

    def on_iteration_begin(self):
        pass

    def on_iteration_end(self):
        self.check_training_status()
        print("Iteration: {} Loss: {:.4f}".format(self.num_iter, self.current_loss.item()))
        self.num_iter += 1

    @torch.no_grad()
    def on_epoch_end(self):
        self.training_loss[1].append(mean(self.training_loss_per_epoch))
        self.training_loss[0].append(self.epoch)
        if (self.epoch % self.frequency_stats == 0) or self.epoch == 0 or (self.epoch == self.num_epochs - 1):
            self.model.eval()
            loss = []
            for data_tensor in self.data_load_loop(self.validation):
                loss = self.on_validation(data_tensor, loss)
            self.model.train()
            val_loss = np.asarray(loss).mean()
            self.validation_loss[1].append(val_loss)
            self.validation_loss[0].append(self.epoch)
            print("\nValidation Loss: {:.4f}\n".format(val_loss))

    @torch.no_grad()
    def on_validation(self, data_tensor, loss):
        output, data = self.model_output(data_tensor)
        loss.append(self.loss(data, output).item())
        return loss

    def on_training_end(self):
        # plot train and val graphs
        plt.plot(self.training_loss[0], self.training_loss[1], 'g', label='Training loss')
        plt.plot(self.validation_loss[0], self.validation_loss[1], 'b', label='validation loss')
        plt.title('Training and Validation loss')
        plt.xlabel('Epochs')
        plt.ylabel('Loss')
        plt.legend()
        plt.show()

    def check_training_status(self):
        self.training_loss_per_epoch.append(self.current_loss.item())
        is_loss_nan = torch.isnan(self.current_loss).item()
        if is_loss_nan:
            self.nan_count += 1
            self.was_previous_loss_nan = True
        else:
            self.nan_count = 0
            self.was_previous_loss_nan = False

        if self.nan_count >= self.max_nans:
            raise ValueError(
                "Loss was NaN {} times".format(self.nan_count)
            )

    def data_load_loop(self, return_set):
        return self._posterior[return_set]

    def train_test_validation(
            self,
            model=None,
            gene_dataset=None,
            type_class=Posterior
    ):
        model = self.model if model is None and hasattr(self, "model") else model
        gene_dataset = self.gene_dataset if gene_dataset is None and hasattr(self, "model") else gene_dataset

        train_size = self.train_size
        test_size = self.test_size
        shuffle = self.shuffle

        n = len(self.gene_dataset)

        if shuffle:
            indices = np.random.permutation(n)
        else:
            indices = np.arange(n)
        train_set = int(np.ceil(n * train_size))
        if test_size is None:
            test_size = 1.0 - train_size
        test_set = int(np.ceil(n * test_size))

        train_indices = indices[:train_set]
        test_indices = indices[train_set:(train_set + test_set)]
        val_indices = indices[(train_set + test_set):]

        if len(val_indices) > 0:
            self.validation = "val_set"
            self.test_on_epoch_end = True
        else:
            self.validation = "test_set"
            self.test_on_epoch_end = False

        return (
            self.create_posterior(
                model, gene_dataset, shuffle=shuffle, indices=train_indices, type_class=type_class
            ),
            self.create_posterior(
                model, gene_dataset, shuffle=shuffle, indices=test_indices, type_class=type_class
            ),
            self.create_posterior(
                model, gene_dataset, shuffle=shuffle, indices=val_indices, type_class=type_class
            )
        )

    def create_posterior(
            self,
            model=None,
            gene_dataset=None,
            shuffle=True,
            indices=None,
            type_class=Posterior
    ):
        model = self.model if model is None and hasattr(self, "model") else model
        gene_dataset = self.gene_dataset if gene_dataset is None and hasattr(self, "model") else gene_dataset

        return type_class(
            model,
            gene_dataset,
            shuffle=shuffle,
            indices=indices,
            use_cuda=self.use_cuda,
            data_loader_kwargs=self.data_loader_kwargs
        )

    def register_posterior(self, train_set, test_set, val_set):
        self._posterior["train_set"] = train_set
        self._posterior["test_set"] = test_set
        self._posterior["val_set"] = val_set

    def save_checkpoint(self, save_dir=None):
        if save_dir is None:
            assert (self.output_dir is not None), 'Please enter an output directory in trainer'
            save_dir = self.output_dir
        # if not os.path.isdir(save_dir):
        #     os.system("mkdir -p {}".format(save_dir))
        torch.save({
            "epoch": self.epoch,
            "model_state_dict": self.model.state_dict(),
            "loss": self.current_loss
        }, save_dir)

    def load_checkpoint(self, saved_file):
        if saved_file is None:
            assert (self.output_dir is not None), 'There is no file specified'
            saved_file = self.output_dir
        if not os.path.isfile(saved_file):
            raise ValueError("File path incorrect. Please specify correct path")
        checkpoint_dict = torch.load(saved_file)
        self.epoch = checkpoint_dict["epoch"]
        self.model.load_state_dict(checkpoint_dict["model_state_dict"])
        self.current_loss = checkpoint_dict["loss"]
        self.model.eval()



import numpy as np
import pandas as pd
import torch
from torch import distributions


def nan2zero(x):
    return torch.where(torch.isnan(x), torch.zeros_like(x), x)


def nan2inf(x):
    return torch.where(torch.isnan(x), torch.zeros_like(x)+np.inf, x)


def _nelem(x):
    nelem = torch.sum(torch.tensor(~torch.isnan(x), dtype=torch.float32))
    return torch.where((torch.eq(nelem, torch.tensor(0.))), torch.tensor(1.), nelem)


def reduce_mean(x):
    nelem = _nelem(x)
    x = nan2zero(x)
    return torch.sum(x) / nelem


# used the proposed implementation as in
# https://discuss.pytorch.org/t/implementing-truncated-normal-initializer/4778/21
def parameterized_truncated_normal(uniform, mu, sigma, a, b):
    normal = distributions.normal.Normal(0, 1)

    if sigma == 0:
        x = torch.zeros(size=uniform.shape)
        x = x + mu
        return x

    alpha = (a - mu) / sigma
    beta = (b - mu) / sigma

    alpha_normal_cdf = normal.cdf(alpha)
    p = alpha_normal_cdf + (normal.cdf(beta) - alpha_normal_cdf) * uniform

    p = p.numpy()
    one = np.array(1, dtype=p.dtype)
    epsilon = np.array(np.finfo(p.dtype).eps, dtype=p.dtype)
    v = np.clip(2 * p - 1, -one + epsilon, one - epsilon)
    x = mu + sigma * np.sqrt(2) * torch.erfinv(torch.from_numpy(v))
    x = torch.clamp(x, a, b)

    return x


def truncated_normal(mean, stddev, shape):
    uniform = torch.from_numpy(np.random.normal(loc=mean, scale=stddev, size=shape))
    return parameterized_truncated_normal(uniform, mu=mean, sigma=stddev, a=-(2*stddev), b=(2*stddev))


